from ... import matchers, utils


@matchers.matcher()
class Any:
    """
    filter any raw text
    """
    type = 'text.Any'

    def __init__(self):
        pass

    def validate(self, message):
        return message.get('data', {}).get('text', {}).get('raw', None)


@matchers.matcher()
class Match:
    type = 'text.Match'

    def __init__(self, test_string):
        self.test_string = test_string

    def validate(self, message):
        return self.test_string == (message.get('data', {}).get('text', {}).get('raw', None))

    def serialize(self):
        return self.test_string

    def deserialize(self, state):
        self.test_string = state

    @staticmethod
    def can_handle(data):
        return utils.is_string(data)

    @staticmethod
    def handle(data):
        return Match(data)



import logging
import pytest
from . import option
from ... import chat, matchers, story
from ...utils import answer, build_fake_session, build_fake_user, SimpleTrigger

logger = logging.getLogger(__name__)


def setup_function(function):
    logger.debug('setup')
    chat.interfaces = {}
    story.stories_library.clear()


@pytest.mark.asyncio
async def test_should_ask_with_options():
    logger.debug('chat.interfaces')
    logger.debug(chat.interfaces)
    session = build_fake_session()
    user = build_fake_user()

    trigger = SimpleTrigger()

    @story.on('How are you?')
    def one_story():
        @story.part()
        async def ask(message):
            return await chat.ask(
                'I feel fine. How about you?',
                options=[{
                    'title': 'Good!',
                    'payload': {'health': 1},
                }, {
                    'title': 'Ugly!',
                    'payload': {'health': -1},
                }, {
                    'title': 'As usual',
                    'payload': {'health': 0},
                }],
                user=message['user'],
            )

        @story.part()
        def get_health(message):
            trigger.receive(message['data']['option'])

    await answer.pure_text('How are you?', session, user)
    await answer.option({'health': 1}, session, user)
    assert trigger.result() == {'health': 1}


@pytest.mark.asyncio
async def test_validate_option():
    session = build_fake_session()
    user = build_fake_user()

    trigger = SimpleTrigger()

    @story.on(receive=option.Any())
    def one_story():
        @story.part()
        def store_option(message):
            trigger.passed()

    await answer.option({'engine': 'start'}, session, user)
    assert trigger.is_triggered


@pytest.mark.asyncio
async def test_validate_only_option():
    session = build_fake_session()
    user = build_fake_user()

    trigger = SimpleTrigger()

    @story.on(receive=option.Any())
    def one_story():
        @story.part()
        def store_option(message):
            trigger.passed()

    await answer.pure_text('Start engine!', session, user)
    assert not trigger.is_triggered


def test_serialize_option_match():
    m_old = option.Match('yellow')
    m_new = matchers.deserialize(matchers.serialize(m_old))
    assert isinstance(m_new, option.Match)
    assert m_new.option == m_old.option


@pytest.mark.asyncio
async def test_validate_only_option():
    session = build_fake_session()
    user = build_fake_user()

    trigger = SimpleTrigger()

    @story.on(receive=option.Match('green'))
    def one_story():
        @story.part()
        def store_option(message):
            trigger.passed()

    await answer.option('green', session, user)
    assert trigger.is_triggered



import pytest
from . import library, parser


@pytest.fixture(scope='function')
def story_library():
    l = library.StoriesLibrary()
    story_1 = parser.ASTNode('hi!')
    story_2 = parser.ASTNode('bye!')
    story_3 = parser.ASTNode('where to go?')
    story_4 = parser.ASTNode('How do you feel?')
    story_1.append(parser.StoryPartFork())
    story_1.add_child(story_4)

    l.add_message_handler(story_1)
    l.add_message_handler(story_2)
    l.add_callable(story_3)
    return l


def test_empty_library_should_return_none():
    lib = library.StoriesLibrary()
    story = lib.get_story_by_topic('hi!')
    assert story is None


def test_get_top_level_story(story_library):
    story = story_library.get_story_by_topic('hi!')
    assert story.topic == 'hi!'


def test_get_substory(story_library):
    stack = [{
        'topic': 'hi!'
    }]
    story = story_library.get_story_by_topic('How do you feel?', stack=stack)
    assert story.topic == 'How do you feel?'



import logging
from unittest import mock
import pytest

from . import messenger
from .. import commonhttp, mockdb, mockhttp
from ... import chat, story, utils
from ...middlewares import option

logger = logging.getLogger(__name__)


def teardown_function(function):
    logger.debug('tear down!')
    story.stories_library.clear()
    chat.interfaces = {}


@pytest.mark.asyncio
async def test_send_text_message():
    user = utils.build_fake_user()

    interface = story.use(messenger.FBInterface(page_access_token='qwerty'))
    mock_http = story.use(mockhttp.MockHttpInterface())

    await interface.send_text_message(
        recipient=user, text='hi!', options=None
    )

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messages/',
        params={
            'access_token': 'qwerty',
        },
        json={
            'message': {
                'text': 'hi!',
            },
            'recipient': {
                'id': user['facebook_user_id'],
            },
        }
    )


@pytest.mark.asyncio
async def test_integration():
    user = utils.build_fake_user()

    story.use(messenger.FBInterface(page_access_token='qwerty'))
    story.use(mockdb.MockDB())
    mock_http = story.use(mockhttp.MockHttpInterface())

    await chat.say('hi there!', user=user)

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messages/',
        params={
            'access_token': 'qwerty',
        },
        json={
            'message': {
                'text': 'hi there!',
            },
            'recipient': {
                'id': user['facebook_user_id'],
            },
        }
    )


@pytest.mark.asyncio
async def test_options():
    user = utils.build_fake_user()

    story.use(messenger.FBInterface(page_access_token='qwerty'))
    story.use(mockdb.MockDB())
    mock_http = story.use(mockhttp.MockHttpInterface())

    await chat.ask(
        'Which color do you like?',
        options=[{
            'title': 'Red',
            'payload': 0xff0000,
        }, {
            'title': 'Green',
            'payload': 0x00ff00,
        }, {
            'title': 'Blue',
            'payload': 0x0000ff,
        }],
        user=user)

    mock_http.post.assert_called_with(
        'https://graph.facebook.com/v2.6/me/messages/',
        params={
            'access_token': 'qwerty',
        },
        json={
            'message': {
                'text': 'Which color do you like?',
                'quick_replies': [
                    {
                        'content_type': 'text',
                        'title': 'Red',
                        'payload': 0xff0000,
                    },
                    {
                        'content_type': 'text',
                        'title': 'Green',
                        'payload': 0x00ff00,
                    },
                    {
                        'content_type': 'text',
                        'title': 'Blue',
                        'payload': 0x0000ff,
                    },
                ],
            },
            'recipient': {
                'id': user['facebook_user_id'],
            },
        }
    )


@pytest.mark.asyncio
async def test_setup_webhook():
    fb_interface = story.use(messenger.FBInterface(
        webhook_url='/webhook',
        webhook_token='some-token',
    ))
    mock_http = story.use(mockhttp.MockHttpInterface())

    mock_http.webhook.assert_called_with(
        '/webhook',
        fb_interface.handle,
        'some-token',
    )


@pytest.mark.asyncio
async def test_should_request_user_data_once_we_do_not_know_current_user():
    fb_interface = story.use(messenger.FBInterface(
        page_access_token='qwerty',
        webhook_url='/webhook',
        webhook_token='some-token',
    ))
    http = story.use(mockhttp.MockHttpInterface(get={
        'first_name': 'Peter',
        'last_name': 'Chang',
        'profile_pic': 'https://fbcdn-profile-a.akamaihd.net/hprofile-ak-xpf1/v/t1.0-1/p200x200/13055603_10105219398495383_8237637584159975445_n.jpg?oh=1d241d4b6d4dac50eaf9bb73288ea192&oe=57AF5C03&__gda__=1470213755_ab17c8c8e3a0a447fed3f272fa2179ce',
        'locale': 'en_US',
        'timezone': -7,
        'gender': 'male'
    }))
    story.use(mockdb.MockDB())

    await fb_interface.handle({
        'object': 'page',
        'entry': [{
            'id': 'PAGE_ID',
            'time': 1473204787206,
            'messaging': [
                {
                    'sender': {
                        'id': 'USER_ID'
                    },
                    'recipient': {
                        'id': 'PAGE_ID'
                    },
                    'timestamp': 1458692752478,
                    'message': {
                        'mid': 'mid.1457764197618:41d102a3e1ae206a38',
                        'seq': 73,
                        'text': 'hello, world!'
                    }
                }
            ]
        }]
    })

    http.get.assert_called_with(
        'https://graph.facebook.com/v2.6/USER_ID',
        params={
            'access_token': 'qwerty',
        },
    )


@pytest.mark.asyncio
async def test_should_request_user_data_and_fail():
    fb_interface = story.use(messenger.FBInterface(
        page_access_token='qwerty',
        webhook_url='/webhook',
        webhook_token='some-token',
    ))
    story.use(mockhttp.MockHttpInterface(
        get_raise=commonhttp.errors.HttpRequestError()))
    db = story.use(mockdb.MockDB())

    await fb_interface.handle({
        'object': 'page',
        'entry': [{
            'id': 'PAGE_ID',
            'time': 1473204787206,
            'messaging': [
                {
                    'sender': {
                        'id': 'USER_ID'
                    },
                    'recipient': {
                        'id': 'PAGE_ID'
                    },
                    'timestamp': 1458692752478,
                    'message': {
                        'mid': 'mid.1457764197618:41d102a3e1ae206a38',
                        'seq': 73,
                        'text': 'hello, world!'
                    }
                }
            ]
        }]
    })

    assert (await db.get_user(facebook_user_id='USER_ID')).no_fb_profile is True


@pytest.mark.asyncio
async def test_webhook_handler_should_return_ok_status_if_http_fail():
    fb_interface = story.use(messenger.FBInterface(
        page_access_token='qwerty',
        webhook_url='/webhook',
        webhook_token='some-token',
    ))
    story.use(mockhttp.MockHttpInterface(get_raise=commonhttp.errors.HttpRequestError()))
    story.use(mockdb.MockDB())

    res = await fb_interface.handle({
        'object': 'page',
        'entry': [{
            'id': 'PAGE_ID',
            'time': 1473204787206,
            'messaging': [
                {
                    'sender': {
                        'id': 'USER_ID'
                    },
                    'recipient': {
                        'id': 'PAGE_ID'
                    },
                    'timestamp': 1458692752478,
                    'message': {
                        'mid': 'mid.1457764197618:41d102a3e1ae206a38',
                        'seq': 73,
                        'text': 'hello, world!'
                    }
                }
            ]
        }]
    })

    assert res['status'] == 200


@pytest.mark.asyncio
async def test_webhook_handler_should_return_ok_status_in_any_case():
    fb_interface = messenger.FBInterface()
    with mock.patch('botstory.integrations.fb.messenger.logger') as mock_logger:
        res = await fb_interface.handle({
            'object': 'page',
            'entry': [{
                'id': 'PAGE_ID',
                'time': 1473204787206,
                'messaging': [
                    {
                        'sender': {
                            'id': 'USER_ID'
                        },
                        'recipient': {
                            'id': 'PAGE_ID'
                        },
                        'timestamp': 1458692752478,
                        'message': {
                            'mid': 'mid.1457764197618:41d102a3e1ae206a38',
                            'seq': 73,
                            'text': 'hello, world!'
                        }
                    }
                ]
            }]
        })

        assert mock_logger.debug.calledWith()

    assert res['status'] == 200

# integration

@pytest.fixture
def build_fb_interface():
    async def builder():
        user = utils.build_fake_user()
        session = utils.build_fake_session()

        storage = story.use(mockdb.MockDB())
        fb = story.use(messenger.FBInterface(page_access_token='qwerty'))

        await storage.set_session(session)
        await storage.set_user(user)

        return fb

    return builder


@pytest.mark.asyncio
async def test_handler_raw_text(build_fb_interface):
    fb_interface = await build_fb_interface()

    correct_trigger = utils.SimpleTrigger()
    incorrect_trigger = utils.SimpleTrigger()

    @story.on('hello, world!')
    def correct_story():
        @story.part()
        def store_result(message):
            correct_trigger.receive(message)

    @story.on('Goodbye, world!')
    def incorrect_story():
        @story.part()
        def store_result(message):
            incorrect_trigger.receive(message)

    await fb_interface.handle({
        'object': 'page',
        'entry': [{
            'id': 'PAGE_ID',
            'time': 1473204787206,
            'messaging': [
                {
                    'sender': {
                        'id': 'USER_ID'
                    },
                    'recipient': {
                        'id': 'PAGE_ID'
                    },
                    'timestamp': 1458692752478,
                    'message': {
                        'mid': 'mid.1457764197618:41d102a3e1ae206a38',
                        'seq': 73,
                        'text': 'hello, world!'
                    }
                }
            ]
        }]
    })

    assert incorrect_trigger.value is None
    assert correct_trigger.value == {
        'user': fb_interface.storage.user,
        'session': fb_interface.storage.session,
        'data': {
            'text': {
                'raw': 'hello, world!'
            }
        }
    }


@pytest.mark.asyncio
async def test_handler_selected_option(build_fb_interface):
    fb_interface = await build_fb_interface()

    correct_trigger = utils.SimpleTrigger()
    incorrect_trigger = utils.SimpleTrigger()

    @story.on(receive=option.Match('GREEN'))
    def correct_story():
        @story.part()
        def store_result(message):
            correct_trigger.receive(message)

    @story.on(receive=option.Match('BLUE'))
    def incorrect_story():
        @story.part()
        def store_result(message):
            incorrect_trigger.receive(message)

    await fb_interface.handle({
        'object': 'page',
        'entry': [{
            'id': 'PAGE_ID',
            'time': 1473204787206,
            'messaging': [{
                'sender': {
                    'id': 'USER_ID'
                },
                'recipient': {
                    'id': 'PAGE_ID'
                },
                'timestamp': 1458692752478,
                'message': {
                    'mid': 'mid.1457764197618:41d102a3e1ae206a38',
                    'seq': 73,
                    'text': 'Green!',
                    'quick_reply': {
                        'payload': 'GREEN'
                    }
                }
            }]
        }]
    })

    assert incorrect_trigger.value is None
    assert correct_trigger.value == {
        'user': fb_interface.storage.user,
        'session': fb_interface.storage.session,
        'data': {
            'option': 'GREEN',
            'text': {
                'raw': 'Green!'
            }
        }
    }



import logging

logger = logging.getLogger(__name__)

from . import callable, parser, processor
from .. import matchers


class Undefined:
    """
    Because we can got ever None value we should have something
    that definitely noted that value wasn't defined
    """

    def __init__(self):
        pass


def match_children(data, key, value):
    step_id = data['stack_tail'][-1]['step']
    fork = data['story'].story_line[step_id]
    if not isinstance(fork, parser.StoryPartFork):
        return []
    return [child for child in fork.children
            if child.extensions.get(key, Undefined) == value]


class Middleware:
    def __init__(self):
        pass

    def process(self, data, validation_result):
        logger.debug('process_switch')
        logger.debug('  data: {}'.format(data))
        logger.debug('  validation_result: {}'.format(validation_result))
        # logger.debug('  children len {}'.format(len(data['story'].children)))
        case_story = match_children(data, 'case_id', validation_result)
        if len(case_story) == 0:
            case_story = match_children(data, 'case_equal', validation_result)
        if len(case_story) == 0:
            case_story = match_children(data, 'default_case', True)

        if len(case_story) == 0:
            logger.debug('   do not have any fork here')
            return data

        logger.debug('  got case_story {}'.format(case_story[0]))

        last_stack_item = data['stack_tail'][-1]
        logger.debug('iterate {} step further'.format(last_stack_item['topic']))
        new_stack_item = {
            'step': last_stack_item['step'] + 1,
            # 'step': last_stack_item['step'],
            'topic': last_stack_item['topic'],
            'data': matchers.serialize(callable.WaitForReturn()),
        }

        return {
            'step': 0,
            'story': case_story[0],
            'stack_tail':
            # data['stack_tail'][:-1] +
                [new_stack_item, processor.build_empty_stack_item()],  # we are going deeper
        }


@matchers.matcher()
class Switch:
    def __init__(self, cases):
        self.cases = cases

    def validate(self, message):
        for case_id, validator in self.cases.items():
            if validator.validate(message):
                return case_id
        return False

    def serialize(self):
        return [{
                    'id': id,
                    'data': matchers.serialize(c),
                } for id, c in self.cases.items()]

    @staticmethod
    def deserialize(data):
        return Switch({
                          case['id']: matchers.deserialize(case['data'])
                          for case in data
                          })


class SwitchOnValue:
    """
    don't need to wait result
    """

    def __init__(self, value):
        self.value = value
        self.immediately = True


async def process_switch_on_value(compiled_story, idx, message, processor, session, waiting_for):
    # ... without waiting for user feedback
    logger.debug('  process immediately')

    waiting_for = await processor.process_next_part_of_story({
        'step': idx,
        'story': compiled_story,
        'stack_tail': [session['stack'].pop()],
    },
        waiting_for.value, session, message,
        bubble_up=False)

    logger.debug('  after process_next_part_of_story')
    logger.debug('      waiting_for = {}'.format(waiting_for))
    logger.debug('      session.stack = {}'.format(session['stack']))

    return waiting_for


class ForkingStoriesAPI:
    def __init__(self, parser_instance):
        self.parser_instance = parser_instance

    def case(self, default=Undefined, equal_to=Undefined, match=Undefined):
        def decorate(story_part):
            compiled_story = self.parser_instance.go_deeper(story_part)
            if default is True:
                compiled_story.extensions['default_case'] = True
            if equal_to is not Undefined:
                compiled_story.extensions['case_equal'] = equal_to
            if match is not Undefined:
                compiled_story.extensions['case_id'] = match
            return story_part

        return decorate



from ... import matchers


@matchers.matcher()
class Any:
    def __init__(self):
        pass

    def validate(self, message):
        return True


@matchers.matcher()
class AnyOf:
    def __init__(self, list_of_matchers):
        self.list_of_matchers = list_of_matchers

    def validate(self, message):
        return len(
            [m for m in self.list_of_matchers if m.validate(message)]
        ) > 0

    def serialize(self):
        return [matchers.serialize(m) for m in self.list_of_matchers]

    @staticmethod
    def deserialize(data):
        return AnyOf([matchers.deserialize(d) for d in data])

    @staticmethod
    def can_handle(data):
        return isinstance(data, list)

    @staticmethod
    def handle(data):
        return AnyOf([matchers.get_validator(r) for r in data])



from aiohttp import test_utils
import json
import pytest
from . import AioHttpInterface
from ..tests.fake_server import fake_fb
from ..commonhttp import errors


@pytest.mark.asyncio
async def test_post(event_loop):
    async with fake_fb.Server(event_loop) as server:
        async with server.session() as session:
            http = AioHttpInterface()
            http.session = session

            assert await http.post(fake_fb.URI.format('/v2.6/me/messages/'), json={'message': 'hello world!'})
            assert len(server.history) == 1
            req = server.history[-1]['request']
            assert req.content_type == 'application/json'
            assert await req.json() == {'message': 'hello world!'}


@pytest.mark.asyncio
async def test_listen_webhook():
    handler = test_utils.make_mocked_coro(return_value={
        'status': 200,
        'content_type': 'application/json',
        'text': json.dumps({'message': 'Ok!'}),
    })
    http = AioHttpInterface(port=9876)
    http.webhook(uri='/webhook', handler=handler, token='qwerty')
    try:
        await http.start()
        res = await http.post_raw('http://localhost:9876/webhook', json={'message': 'Is there anybody in there?'})
        handler.assert_called_once_with({'message': 'Is there anybody in there?'})
        assert res['status'] == 200
        assert res['text'] == json.dumps({'message': 'Ok!'})
    finally:
        await http.stop()


@pytest.mark.asyncio
async def test_pass_validation_for_correct_request():
    http = AioHttpInterface(port=9876)
    http.webhook(uri='/webhook', handler=None, token='token-value')
    try:
        await http.start()
        res = await http.get('http://localhost:9876/webhook', params={
            'hub.challenge': 'some-challenge',
            'hub.mode': 'subscribe',
            'hub.verify_token': 'token-value',
        })
        assert res == 'some-challenge'
    finally:
        await http.stop()


@pytest.mark.asyncio
async def test_reject_validation_for_incorrect_request():
    http = AioHttpInterface(port=9876)
    http.webhook(uri='/webhook', handler=None, token='token-value')
    try:
        await http.start()
        await http.get('http://localhost:9876/webhook', params={
            'something': 'incorrect',
        })
    except errors.HttpRequestError as err:
        assert err.code == 422
        assert err.message == 'Error, wrong validation token'
    finally:
        await http.stop()


@pytest.mark.asyncio
async def test_should_not_create_server_if_there_wasnt_any_webhooks():
    http = AioHttpInterface(port=9876)
    try:
        await http.start()
        assert not http.server
    finally:
        await http.stop()


@pytest.mark.skip(reason='too long')
@pytest.mark.asyncio
async def test_get_400_from_wrong_domain():
    http = AioHttpInterface()
    try:
        await http.get('http://wrong-url')
    except errors.HttpRequestError as err:
        assert err.code == 400


@pytest.mark.asyncio
async def test_get_400_from_wrong_path():
    http = AioHttpInterface()
    try:
        await http.get('http://localhost:9876/webhook')
    except errors.HttpRequestError as err:
        assert err.code == 400


@pytest.mark.asyncio
async def test_post_to_wrong_path_get_400(event_loop):
    async with fake_fb.Server(event_loop) as server:
        async with server.session() as session:
            http = AioHttpInterface()
            http.session = session

            try:
                await http.post(fake_fb.URI.format('/v2.6/me/messages/mistake'), json={'message': 'hello world!'})
            except errors.HttpRequestError as err:
                assert err.code == 404


@pytest.mark.skip(reason='too long')
@pytest.mark.asyncio
async def test_post_400_from_wrong_domain():
    http = AioHttpInterface()
    try:
        await http.post('http://wrong-url')
    except errors.HttpRequestError as err:
        assert err.code == 400


@pytest.mark.asyncio
async def test_post_400_from_wrong_path():
    http = AioHttpInterface()
    try:
        await http.post('http://localhost:9876/webhook')
    except errors.HttpRequestError as err:
        assert err.code == 400



import logging
import pytest
import random

from .. import chat, story
from ..utils import answer, build_fake_session, build_fake_user, SimpleTrigger

logger = logging.getLogger(__name__)


def teardown_function(function):
    logger.debug('tear down!')
    story.stories_library.clear()


@pytest.mark.asyncio
async def test_begin_of_callable_story():
    trigger = SimpleTrigger()
    session = build_fake_session()

    @story.callable()
    def one_story():
        @story.part()
        def store_arguments(arg1, arg2):
            trigger.receive({
                'value1': arg1,
                'value2': arg2,
            })

    await one_story(arg1=1, arg2=2, session=session)

    assert trigger.result() == {
        'value1': 1,
        'value2': 2,
    }


@pytest.mark.asyncio
async def test_parts_of_callable_story():
    trigger_1 = SimpleTrigger()
    trigger_2 = SimpleTrigger()
    session = build_fake_session()
    user = build_fake_user()

    @story.callable()
    def meet_ava_story():
        @story.part()
        async def ask_name(user):
            return await chat.ask(
                'My name is Ava. What is your name?',
                user=user,
            )

        @story.part()
        async def ask_age(message):
            trigger_1.passed()
            return await chat.ask(
                'Nice to see you {}. What do you do here?'.format(message['data']['text']['raw']),
                user=message['user'],
            )

        @story.part()
        async def store_arguments(message):
            age = int(message['data']['text']['raw'])
            if age < 30:
                res = 'You are so young! '
            else:
                res = 'Hm. Too old to die young'

            await chat.say(res, user=message['user'])
            trigger_2.passed()

    await meet_ava_story(user, session=session)

    await answer.pure_text('Eugene', session, user=user)
    await answer.pure_text('13', session, user=user)

    assert trigger_1.is_triggered
    assert trigger_2.is_triggered

@pytest.mark.asyncio
async def test_call_story_from_common_story():
    trigger = SimpleTrigger()

    session = build_fake_session()
    user = build_fake_user()

    @story.callable()
    def common_greeting():
        @story.part()
        async def ask_name(user):
            return await chat.ask(
                'Hi {}. How are you?'.format(user['name']),
                user=user,
            )

    @story.on('Hi!')
    def meet():
        @story.part()
        async def greeting(message):
            return await common_greeting(
                user=message['user'],
                session=message['session'],
            )

        @story.part()
        async def ask_location(message):
            return await chat.ask(
                'Which planet are we going to visit today?',
                user=message['user'],
            )

        @story.part()
        def parse(message):
            trigger.receive(message['data']['text']['raw'])

    await answer.pure_text('Hi!', session, user=user)
    await answer.pure_text('I\'m fine', session, user=user)
    await answer.pure_text('Venus, as usual!', session, user=user)

    assert trigger.value == 'Venus, as usual!'


@pytest.mark.asyncio
async def test_parts_of_callable_story_can_be_sync():
    trigger_1 = SimpleTrigger()
    trigger_2 = SimpleTrigger()
    session = build_fake_session()

    @story.callable()
    def one_story():
        @story.part()
        def has():
            trigger_1.passed()

        @story.part()
        def so():
            trigger_2.passed()

    await one_story(session=session)

    assert trigger_1.is_triggered
    assert trigger_2.is_triggered


@pytest.mark.asyncio
async def test_call_story_from_another_callable():
    trigger_1 = SimpleTrigger()
    trigger_2 = SimpleTrigger()
    session = build_fake_session()

    @story.callable()
    def one_story():
        @story.part()
        def so_1(session_1):
            pass

        @story.part()
        async def so_2(session_2):
            await another_story(session=session_2)

        @story.part()
        def so_3(session_3):
            trigger_2.passed()

    @story.callable()
    def another_story():
        @story.part()
        def has():
            pass

        @story.part()
        def so():
            trigger_1.passed()

    # push extra parameter with session
    # and it will propagate up to other story as well
    await one_story(session, session=session)

    assert trigger_1.is_triggered
    assert trigger_2.is_triggered


@pytest.mark.asyncio
async def test_async_end_of_story():
    sides = ['heads', 'tails']
    user = build_fake_user()
    session = build_fake_session()
    budget = SimpleTrigger(2)
    game_result = SimpleTrigger()

    @story.callable()
    def flip_a_coin():
        @story.part()
        async def choose_side(user):
            return await chat.ask('Please choose a side', user=user)

        @story.part()
        async def flip(message):
            user_side = message['data']['text']['raw']
            await chat.say('Thanks!', user=message['user'])
            await chat.say('And I am flipping a Coin', user=message['user'])
            coin_side = random.choice(sides)
            await chat.say('and got {}'.format(coin_side), user=message['user'])
            if coin_side == user_side:
                await chat.say('My greetings! You guessed!', user=message['user'])
                budget.receive(budget.value + 1)
            else:
                await chat.say('Sad but you loose!', user=message['user'])
                budget.receive(budget.value - 1)
            return story.SwitchOnValue(budget.value)

        @story.case(equal_to=0)
        def loose():
            @story.part()
            def end_of_story(message):
                return story.EndOfStory({
                    'game_result': 'loose'
                })

        @story.case(equal_to=5)
        def win():
            @story.part()
            def end_of_story(message):
                return story.EndOfStory({
                    'game_result': 'win'
                })

        @story.part()
        def tail_recursion(message):
            # TODO: add test for recursion
            # return flip_a_coin(message['user'], session)
            return story.EndOfStory({
                'game_result': 'in progress'
            })

    @story.on('enter to the saloon')
    def enter_to_the_saloon():
        @story.part()
        async def start_a_game(message):
            return await flip_a_coin(user, session=message['session'])

        @story.part()
        def game_over(message):
            logger.debug('game_over')
            logger.debug(message)
            game_result.receive(message['data']['game_result'])

    await answer.pure_text('enter to the saloon',
                     session=session, user=user)

    await answer.pure_text(random.choice(sides), session, user)

    assert game_result.result() in ['loose', 'win', 'in progress']


@pytest.mark.asyncio
async def test_sync_end_of_story():
    session = build_fake_session()

    part_1 = SimpleTrigger()
    part_2 = SimpleTrigger()
    part_3 = SimpleTrigger()

    @story.callable()
    def one_story():
        @story.part()
        def story_part_1():
            part_1.passed()

        @story.part()
        def story_part_2():
            part_2.passed()
            return story.EndOfStory('Break Point')

        @story.part()
        def story_part_3():
            part_3.passed()

    logger.debug('one_story')
    logger.debug(one_story)
    res = await one_story(session=session)

    assert part_1.is_triggered
    assert part_2.is_triggered
    assert not part_3.is_triggered
    assert res == 'Break Point'



import logging
from .. import commonhttp

logger = logging.getLogger(__name__)


class FBInterface:
    type = 'interface.facebook'

    def __init__(self,
                 api_uri='https://graph.facebook.com/v2.6',
                 page_access_token=None,
                 webhook_url=None,
                 webhook_token=None,
                 ):
        """

        :param api_uri:
        :param page_access_token:
        :param webhook_url:
        :param webhook_token:
        """
        self.api_uri = api_uri
        self.token = page_access_token
        self.webhook = webhook_url
        self.webhook_token = webhook_token

        self.http = None
        self.processor = None
        self.storage = None

    async def send_text_message(self, recipient, text, options=None):
        """
        async send message to the facebook user (recipient)

        :param session:
        :param recipient:
        :param text:
        :param options:

        :return:
        """

        if not options:
            options = []

        message = {
            'text': text,
        }

        quick_replies = [{**reply, 'content_type': 'text'} for reply in options]
        if len(quick_replies) > 0:
            message['quick_replies'] = quick_replies

        return await self.http.post(
            self.api_uri + '/me/messages/',
            params={
                'access_token': self.token,
            },
            json={
                'recipient': {
                    'id': recipient['facebook_user_id'],
                },
                'message': message,
            })

    def add_http(self, http):
        logger.debug('add_http')
        logger.debug(http)
        self.http = http
        if self.webhook:
            http.webhook(self.webhook, self.handle, self.webhook_token)

    def add_storage(self, storage):
        logger.debug('add_storage')
        logger.debug(storage)
        self.storage = storage

    async def request_profile(self, facebook_user_id):
        """
        Make request to facebook
        to receive more information about user

        More: https://developers.facebook.com/docs/messenger-platform/user-profile

        :param facebook_user_id:
        :return:
        """
        return await self.http.get(
            '{}/{}'.format(self.api_uri, facebook_user_id),
            params={
                'access_token': self.token,
            },
        )

    async def handle(self, data):
        logger.debug('')
        logger.debug('> handle <')
        logger.debug('')
        logger.debug('  entry: {}'.format(data))
        try:
            for e in data.get('entry', []):
                messaging = e.get('messaging', [])
                logger.debug('  messaging: {}'.format(messaging))

                if len(messaging) == 0:
                    logger.warning('  entry {} list lack of "messaging" field'.format(e))

                for m in messaging:
                    logger.debug('  m: {}'.format(m))

                    facebook_user_id = m['sender']['id']

                    user = await self.storage.get_user(facebook_user_id=facebook_user_id)
                    if not user:
                        logger.debug('  should create new user {}'.format(facebook_user_id))

                        try:
                            messenger_profile_data = await self.request_profile(facebook_user_id)
                        except commonhttp.errors.HttpRequestError as err:
                            messenger_profile_data = {
                                'no_fb_profile': True,
                            }

                        user = await self.storage.new_user(
                            facebook_user_id=facebook_user_id,
                            no_fb_profile=messenger_profile_data.get('no_fb_profile', None),
                            first_name=messenger_profile_data.get('first_name', None),
                            last_name=messenger_profile_data.get('last_name', None),
                            profile_pic=messenger_profile_data.get('profile_pic', None),
                            locale=messenger_profile_data.get('locale', None),
                            timezone=messenger_profile_data.get('timezone', None),
                            gender=messenger_profile_data.get('gender', None),
                        )

                    session = await self.storage.get_session(facebook_user_id=facebook_user_id)
                    if not session:
                        logger.debug('  should create new session for user {}'.format(facebook_user_id))
                        session = await self.storage.new_session(
                            facebook_user_id=facebook_user_id,
                            stack=[],
                            user=user,
                        )

                    message = {
                        'session': session,
                        'user': user,
                    }
                    raw_message = m.get('message', {})

                    if raw_message == {}:
                        logger.warning('  entry {} "message" field is empty'.format(e))

                    logger.debug('  raw_message: {}'.format(raw_message))

                    data = {}
                    text = raw_message.get('text', None)
                    if text is not None:
                        data['text'] = {
                            'raw': text,
                        }
                    else:
                        logger.warning('  entry {} "text"'.format(e))

                    quick_reply = raw_message.get('quick_reply', None)
                    if quick_reply is not None:
                        data['option'] = quick_reply['payload']

                    message['data'] = data

                    await self.processor.match_message(message)
        except BaseException as err:
            # logger.exception(err)
            pass

        return {
            'status': 200,
            'text': 'Ok!',
        }



import logging
import os
import pytest

from . import fake_server
from .. import fb, aiohttp, mongodb, mockhttp
from ... import story, chat, utils

logger = logging.getLogger(__name__)


def teardown_function(function):
    logger.debug('tear down!')
    story.stories_library.clear()


@pytest.fixture
def build_context():
    async def builder(db, no_session=False, no_user=False):
        user = None
        if not no_user:
            user = utils.build_fake_user()
            await db.set_user(user)

        if not no_session:
            session = utils.build_fake_session(user=user)
            await db.set_session(session)

        story.use(db)
        interface = story.use(fb.FBInterface(page_access_token='qwerty'))
        story.use(mockhttp.MockHttpInterface())

        return interface, user

    return builder


@pytest.fixture
@pytest.mark.asyncio
def open_db():
    class AsyncDBConnection:
        def __init__(self):
            self.db_interface = mongodb.MongodbInterface(uri=os.environ.get('TEST_MONGODB_URL', 'mongo'),
                                                         db_name='test')

        async def __aenter__(self):
            await self.db_interface.start()
            await self.db_interface.clear_collections()
            return self.db_interface

        async def __aexit__(self, exc_type, exc_val, exc_tb):
            self.db_interface.stop()
            await self.db_interface.clear_collections()
            self.db_interface = None

    return AsyncDBConnection


@pytest.mark.asyncio
async def test_facebook_interface_should_use_aiohttp_to_post_message(event_loop):
    async with fake_server.fake_fb.Server(event_loop) as server:
        async with server.session() as server_session:
            # 1) setup app

            try:
                story.use(fb.FBInterface(
                    webhook_url='/webhook',
                ))
                http_integration = story.use(aiohttp.AioHttpInterface())

                await story.start()

                # 2) and test it

                # TODO: could be better. Mock http interface
                http_integration.session = server_session
                user = utils.build_fake_user()

                # 3) test

                await chat.say('Pryvit!', user=user)

                assert len(server.history) > 0
                req = server.history[-1]['request']
                assert req.content_type == 'application/json'
                assert await req.json() == {
                    'recipient': {
                        'id': user['facebook_user_id'],
                    },
                    'message': {
                        'text': 'Pryvit!'
                    }
                }
            finally:
                await story.stop()


@pytest.mark.asyncio
async def test_integrate_mongodb_with_facebook(open_db, build_context):
    async with open_db() as mongodb:
        facebook, user = await build_context(mongodb)

        trigger = utils.SimpleTrigger()

        @story.on('hello, world!')
        def correct_story():
            @story.part()
            def store_result(message):
                trigger.receive(message)

        await facebook.handle({
            "object": "page",
            "entry": [{
                'id': 'PAGE_ID',
                'time': 1473204787206,
                'messaging': [{
                    'sender': {
                        'id': user['facebook_user_id'],
                    },
                    'recipient': {
                        'id': 'PAGE_ID'
                    },
                    'timestamp': 1458692752478,
                    'message': {
                        'mid': 'mid.1457764197618:41d102a3e1ae206a38',
                        'seq': 73,
                        'text': 'hello, world!'
                    }
                }
                ]
            }]
        })

        del trigger.value['session']
        assert trigger.value == {
            'user': user,
            'data': {
                'text': {
                    'raw': 'hello, world!'
                }
            }
        }


@pytest.mark.asyncio
async def test_integrate_mongodb_with_facebook_with_none_session(open_db, build_context):
    async with open_db() as mongodb:
        facebook, _ = await build_context(mongodb, no_session=True, no_user=True)

        trigger = utils.SimpleTrigger()

        @story.on('hello, world!')
        def correct_story():
            @story.part()
            def store_result_for_new_user(message):
                trigger.receive(message)

        await facebook.handle({
            "object": "page",
            "entry": [{
                'id': 'PAGE_ID',
                'time': 1473204787206,
                'messaging': [{
                    'sender': {
                        'id': 'some-facebook-id',
                    },
                    'recipient': {
                        'id': 'PAGE_ID'
                    },
                    'timestamp': 1458692752478,
                    'message': {
                        'mid': 'mid.1457764197618:41d102a3e1ae206a38',
                        'seq': 73,
                        'text': 'hello, world!'
                    }
                }]
            }]
        })

        assert trigger.value
        assert trigger.value['data']['text']['raw'] == 'hello, world!'
        assert trigger.value['user']['facebook_user_id'] == 'some-facebook-id'



import json
import logging
import pytest
import random

logger = logging.getLogger(__name__)

from . import forking
from .. import chat, story, matchers
from ..middlewares import location, text
from ..utils import answer, build_fake_session, build_fake_user, SimpleTrigger


def teardown_function(function):
    logger.debug('tear down!')
    story.stories_library.clear()


@pytest.mark.asyncio
async def test_cases():
    user = build_fake_user()
    session = build_fake_session()

    trigger_location = SimpleTrigger()
    trigger_text = SimpleTrigger()
    trigger_after_switch = SimpleTrigger()

    @story.on('Hi there!')
    def one_story():
        @story.part()
        async def start(message):
            await chat.say('Where do you go?', user=message['user'])
            return forking.Switch({
                'location': location.Any(),
                'text': text.Any(),
            })

        @story.case(match='location')
        def location_case():
            @story.part()
            def store_location(message):
                trigger_location.receive(message['data']['location'])

        @story.case(match='text')
        def text_case():
            @story.part()
            def store_location(message):
                trigger_text.receive(message['data']['text']['raw'])

        @story.part()
        def after_switch(message):
            trigger_after_switch.passed()

    await answer.pure_text('Hi there!', session, user)
    await answer.location({'x': 123, 'y': 321}, session, user)

    assert trigger_location.result() == {'x': 123, 'y': 321}
    assert not trigger_text.result()
    assert trigger_after_switch.is_triggered


@pytest.mark.asyncio
async def test_sync_value():
    user = build_fake_user()
    session = build_fake_session()
    trigger_start = SimpleTrigger()
    trigger_heads = SimpleTrigger()
    trigger_tails = SimpleTrigger()

    @story.on('Flip a coin!')
    def one_story():
        @story.part()
        def start(message):
            coin = random.choice(['heads', 'tails'])
            assert not trigger_start.is_triggered
            assert not trigger_heads.is_triggered
            assert not trigger_tails.is_triggered
            trigger_start.passed()
            return forking.SwitchOnValue(coin)

        @story.case(equal_to='heads')
        def heads():
            @story.part()
            def store_heads(message):
                assert not trigger_heads.is_triggered
                trigger_heads.passed()

        @story.case(equal_to='tails')
        def tails():
            @story.part()
            def store_tails(message):
                assert not trigger_tails.is_triggered
                trigger_tails.passed()

    await answer.pure_text('Flip a coin!', session, user)

    assert trigger_heads.is_triggered != trigger_tails.is_triggered


@pytest.mark.asyncio
async def test_few_switches_in_one_story():
    user = build_fake_user()
    session = build_fake_session()
    trigger_heads = SimpleTrigger()
    trigger_heads.receive(0)
    trigger_tails = SimpleTrigger()
    trigger_tails.receive(0)

    @story.on('Flip a coin!')
    def one_story():
        @story.part()
        def start(message):
            coin = random.choice(['heads', 'tails'])
            return forking.SwitchOnValue(coin)

        @story.case(equal_to='heads')
        def heads_0():
            @story.part()
            def store_heads(message):
                trigger_heads.receive(trigger_heads.value + 1)

        @story.case(equal_to='tails')
        def tails_0():
            @story.part()
            def store_tails(message):
                trigger_tails.receive(trigger_tails.value + 1)

        @story.part()
        def start(message):
            assert trigger_heads.value + trigger_tails.value == 1
            coin = random.choice(['heads', 'tails'])
            return forking.SwitchOnValue(coin)

        @story.case(equal_to='heads')
        def heads_1():
            @story.part()
            def store_heads(message):
                trigger_heads.receive(trigger_heads.value + 1)

        @story.case(equal_to='tails')
        def tails_1():
            @story.part()
            def store_tails(message):
                trigger_tails.receive(trigger_tails.value + 1)

    await answer.pure_text('Flip a coin!', session, user)

    assert trigger_heads.value + trigger_tails.value == 2


@pytest.mark.asyncio
async def test_default_sync_value():
    user = build_fake_user()
    session = build_fake_session()
    trigger_1 = SimpleTrigger()
    trigger_default = SimpleTrigger()

    @story.on('Roll the dice!')
    def one_story_1():
        @story.part()
        def start(message):
            side = random.randint(1, 6)
            return forking.SwitchOnValue(side)

        @story.case(equal_to=1)
        def side_1():
            @story.part()
            def store_1(message):
                trigger_1.passed()

        @story.case(default=True)
        def other_sides():
            @story.part()
            def store_other(message):
                trigger_default.passed()

    await answer.pure_text('Roll the dice!', session, user)

    assert trigger_1.is_triggered != trigger_default.is_triggered


@pytest.mark.asyncio
async def test_one_sync_switch_inside_of_another_sync_switch():
    user = build_fake_user()
    session = build_fake_session()
    visited_rooms = SimpleTrigger(0)

    @story.on('enter')
    def labyrinth():
        @story.part()
        def enter(message):
            turn = random.choice(['left', 'right'])
            visited_rooms.receive(visited_rooms.value + 1)
            return forking.SwitchOnValue(turn)

        @story.case(equal_to='left')
        def room_1():
            @story.part()
            def next_room_1(message):
                turn = random.choice(['left', 'right'])
                visited_rooms.receive(visited_rooms.value + 1)
                return forking.SwitchOnValue(turn)

            @story.case(equal_to='left')
            def room_1_1():
                @story.part()
                def next_room_1_1(message):
                    visited_rooms.receive(visited_rooms.value + 1)

            @story.case(equal_to='right')
            def room_1_2():
                @story.part()
                def next_room_1_2(message):
                    visited_rooms.receive(visited_rooms.value + 1)

        @story.case(equal_to='right')
        def room_2():
            @story.part()
            def next_room_2(message):
                turn = random.choice(['left', 'right'])
                visited_rooms.receive(visited_rooms.value + 1)
                return forking.SwitchOnValue(turn)

            @story.case(equal_to='left')
            def room_2_1():
                @story.part()
                def next_room_2_1(message):
                    visited_rooms.receive(visited_rooms.value + 1)

            @story.case(equal_to='right')
            def room_2_2():
                @story.part()
                def next_room_2_2(message):
                    visited_rooms.receive(visited_rooms.value + 1)

    await answer.pure_text('enter', session, user)

    assert visited_rooms.value == 3


@pytest.mark.asyncio
async def test_one_sync_switch_inside_of_another_async_switch():
    user = build_fake_user()
    session = build_fake_session()
    visited_rooms = SimpleTrigger(0)

    @story.on('enter')
    def labyrinth():
        @story.part()
        async def enter(message):
            visited_rooms.receive(visited_rooms.value + 1)
            return await chat.ask('Which turn to choose?', user=message['user'])

        @story.part()
        def parse_direction_0(message):
            return forking.SwitchOnValue(message['data']['text']['raw'])

        @story.case(equal_to='left')
        def room_1():
            @story.part()
            async def next_room_1(message):
                visited_rooms.receive(visited_rooms.value + 1)
                return await chat.ask('Which turn to choose?', user=message['user'])

            @story.part()
            def parse_direction_1(message):
                return forking.SwitchOnValue(message['data']['text']['raw'])

            @story.case(equal_to='left')
            def room_1_1():
                @story.part()
                def next_room_1_1(message):
                    visited_rooms.receive(visited_rooms.value + 1)

            @story.case(equal_to='right')
            def room_1_2():
                @story.part()
                def next_room_1_2(message):
                    visited_rooms.receive(visited_rooms.value + 1)

        @story.case(equal_to='right')
        def room_2():
            @story.part()
            async def next_room_2(message):
                visited_rooms.receive(visited_rooms.value + 1)
                return await chat.ask('Which turn to choose?', user=message['user'])

            @story.part()
            def parse_direction_2(message):
                return forking.SwitchOnValue(message['data']['text']['raw'])

            @story.case(equal_to='left')
            def room_2_1():
                @story.part()
                def next_room_2_1(message):
                    visited_rooms.receive(visited_rooms.value + 1)

            @story.case(equal_to='right')
            def room_2_2():
                @story.part()
                def next_room_2_2(message):
                    visited_rooms.receive(visited_rooms.value + 1)

    await answer.pure_text('enter', session, user)
    await answer.pure_text(random.choice(['left', 'right']), session, user)
    await answer.pure_text(random.choice(['left', 'right']), session, user)

    assert visited_rooms.value == 3


@pytest.mark.asyncio
async def test_switch_inside_of_callable_inside_of_switch():
    user = build_fake_user()
    session = build_fake_session()

    visited_rooms = SimpleTrigger(0)
    spell_type = SimpleTrigger()
    spell_power = SimpleTrigger()

    @story.callable()
    def cast_the_magic():
        @story.part()
        async def ask_kind_of_spell(user):
            return await chat.ask('What kind of spell do you cast?', user=user)

        @story.part()
        def switch_by_kind_of_spell(message):
            return forking.SwitchOnValue(message['data']['text']['raw'])

        @story.case(equal_to='fireball')
        def fireball():
            @story.part()
            async def power_of_spell(message):
                spell_type.receive(message['data']['text']['raw'])
                return await chat.ask('What is the power of fireball?', user=message['user'])

        @story.case(equal_to='lightning')
        def lightning():
            @story.part()
            async def power_of_spell(message):
                spell_type.receive(message['data']['text']['raw'])
                return await chat.ask('What is the power of lightning?', user=message['user'])

        @story.part()
        def store_power(message):
            spell_power.receive(message['data']['text']['raw'])

    @story.on('enter')
    def dungeon():
        @story.part()
        async def ask_direction(message):
            return await chat.ask(
                'Where do you go?',
                user=message['user']
            )

        @story.part()
        def parser_direction(message):
            return forking.SwitchOnValue(message['data']['text']['raw'])

        @story.case(equal_to='left')
        def room_1():
            @story.part()
            async def meet_dragon(message):
                return await cast_the_magic(message['user'], session=session)

            @story.part()
            def store_end(message):
                visited_rooms.receive(visited_rooms.value + 1)

        @story.case(equal_to='right')
        def room_2():
            @story.part()
            async def meet_ogr(message):
                return await cast_the_magic(message['user'], session=session)

            @story.part()
            def store_end(message):
                visited_rooms.receive(visited_rooms.value + 1)

    await answer.pure_text('enter', session, user)
    await answer.pure_text(random.choice(['left', 'right']), session, user)
    await answer.pure_text(random.choice(['fireball', 'lightning']), session, user)
    await answer.pure_text(random.choice(['light', 'strong']), session, user)

    assert visited_rooms.value == 1
    assert spell_type.value in ['fireball', 'lightning']
    assert spell_power.value in ['light', 'strong']


@pytest.mark.asyncio
async def test_switch_without_right_case():
    user = build_fake_user()
    session = build_fake_session()

    get_help = SimpleTrigger()
    say_goodbay = SimpleTrigger()

    @story.on('I do not know')
    def meet_someone():
        @story.part()
        async def ask(message):
            return await chat.ask(
                'Do you need a help?',
                user=message['user'],
            )

        @story.part()
        def parse_result(message):
            return forking.SwitchOnValue(message['data']['text']['raw'])

        @story.case(equal_to='yes')
        def yes():
            @story.part()
            async def lets_go(message):
                await chat.say('Let\'s google together!', message['user'])
                get_help.passed()

        @story.part()
        async def see_you(message):
            await chat.say('Nice to see you!', user=message['user'])
            say_goodbay.passed()

    await answer.pure_text('I do not know', session, user)
    await answer.pure_text('No', session, user)

    assert not get_help.is_triggered
    assert say_goodbay.is_triggered


def test_serialize():
    m_old = forking.Switch({
        'location': location.Any(),
        'text': text.Any(),
    })
    ready_to_store = json.dumps(matchers.serialize(m_old))
    print('ready_to_store', ready_to_store)
    m_new = matchers.deserialize(json.loads(ready_to_store))

    assert isinstance(m_new, forking.Switch)
    assert isinstance(m_new.cases['location'], type(m_old.cases['location']))
    assert isinstance(m_new.cases['text'], type(m_old.cases['text']))



from unittest import mock
import aiohttp
from aiohttp.helpers import sentinel

def stub(name=None):
    """
    Creates a stub method. It accepts any arguments. Ideal to register to
    callbacks in tests.
    :param name: the constructed stub's name as used in repr
    :rtype: mock.MagicMock
    :return: Stub object.
    """
    return mock.MagicMock(spec=lambda *args, **kwargs: None, name=name)


class MockHttpInterface:
    def __init__(self, get={}, get_raise=sentinel,
                 post=True, post_raise=sentinel,
                 start=True,
                 stop=True):
        self.get = aiohttp.test_utils.make_mocked_coro(return_value=get, raise_exception=get_raise)
        self.post = aiohttp.test_utils.make_mocked_coro(return_value=post, raise_exception=post_raise)
        self.start = aiohttp.test_utils.make_mocked_coro(return_value=start)
        self.stop = aiohttp.test_utils.make_mocked_coro(return_value=stop)
        self.webhook = stub('webhook')



import pytest
from . import location
from ... import matchers, story
from ...utils import answer, build_fake_session, build_fake_user, SimpleTrigger


def teardown_function(function):
    print('tear down!')
    story.stories_library.clear()


@pytest.mark.asyncio
async def test_should_trigger_on_any_location():
    trigger = SimpleTrigger()
    session = build_fake_session()
    user = build_fake_user()

    @story.on(receive=location.Any())
    def one_story():
        @story.part()
        def then(message):
            trigger.passed()

    await answer.location({'lat': 1, 'lng': 1}, session, user)
    assert trigger.is_triggered


@pytest.mark.asyncio
async def test_should_not_react_on_common_message():
    trigger = SimpleTrigger()
    session = build_fake_session()
    user = build_fake_user()

    @story.on(receive=location.Any())
    def one_story():
        @story.part()
        def then(message):
            trigger.passed()

    await answer.pure_text('Hey!', session, user)

    assert not trigger.is_triggered


def test_serialize_location():
    m_old = location.Any()
    m_new = matchers.deserialize(matchers.serialize(m_old))
    assert isinstance(m_new, location.Any)



from .. import matchers


class CommonStoriesAPI:
    def __init__(self, parser_instance, library):
        self.parser_instance = parser_instance
        self.library = library

    def on(self, receive):
        def fn(one_story):
            compiled_story = self.parser_instance.compile(
                one_story,
            )
            compiled_story.extensions['validator'] = matchers.get_validator(receive)
            self.library.add_message_handler(compiled_story)

            return one_story

        return fn

    def part(self):
        def fn(part_of_story):
            self.parser_instance.part(part_of_story)
            return part_of_story

        return fn
